/* VC_DAD_ENV_101.c
 *  
 * 
 * Device check default values:
 * avrdude -p m328p -c avrisp2                   						; command
 * avrdude: Device signature = 0x1e950f   (probably m328p)	; default values for m328p
 * avrdude: safemode: Fuses OK (E:FF, H:D9, L:62) 					; default values for m328p
 * 
 * 
 * Fuse setings for 20MHz external Quarz:
 * - avrdude -p m328p -c avrisp2 -U lfuse:w:0xe7:m
 * - avrdude -p m328p -c avrisp2 -U hfuse:w:0xd9:m
 * 
 * 
 * compile and flash:
 * - avr-gcc -Wall -Os -mmcu=atmega328p VC_DAD_ENV_101.c -o VC_DAD_ENV_101.elf
 * - avr-objcopy -O ihex -j .text -j .data VC_DAD_ENV_101.elf VC_DAD_ENV_101.hex
 * - avrdude -p m328p -c avrisp2 -U flash:w:VC_DAD_ENV_101.hex
 * 
 * 
 * Hardware: m328p at 20Mhz external
 * Board DAR_ENV V-1.0.1-A+B / 2022_11
 * 
 * ================================================================================
 * 
 * Ports used:
 * PB2-PB5 SPI interface
 * 
 * PC0 Delay CV
 * PC1 Attack CV
 * PC2 Decay CV
 * 
 * 
 * PD0 LED/Out Decay
 * PD1 LED/Out Attack
 * PD2 INT 0
 * PD3 INT 1
 * PD5 LED/Out Delay
 * PD6 LED/Out EOC 
 * PD7 Gate/In
 * 
 * =========================================================================================
 * 
 * Generates a delay/attack/decay envelope on rising edge of gate. Further gates ignored until end of
 * decay is finished (EOC). Delay/attack/decay time is manually and voltage controllable. Gate is polled every loop.
 * CV input parameters are read with timer0 interrupt ISR (TIMER0_OVF_vect). 
 * Delay-, attack-, delay- and EOC LED and outputs are switched.
 * 
 * xx.12.2022
 * Working
 * 
 * 01.06.2023
 * Adapted to updated hardware. Board V 1.1.0-B. Gate input moved from PD3 to PD7
 * Routines for ISR(INT0_vect) and ISR(INT1_vect) implemented. Yet empty.
 * 
 * 11.03.2024
 * Second lookup table implemented for different timing in attack and decay section
 * Tables moved to flash. (PROGMEM)
 * 
*/

#define config_bits 0x03		// Mode set for MCP4921; output gain selected 1x
#define F_CPU 20000000			// Used in delay.h

#include <avr/io.h>
#include <avr/interrupt.h>
#include <math.h>
#include <stdbool.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
																							// Tables loaded to flash (PROGMEM)
#include "lookup_attack_DAD_7_4_1mS_5S.c"			// Mapping input voltage CVA to attack increment. Determine the attack timing.
#include "lookup_decay_DAD_7_4_1mS_5S.c"			// Mapping input voltage CVD to decay increment. Determine the decay timing.

volatile float attack_time;										// Must be float here because of the calculation (no casting)
volatile float decay_time;										// Must be float here because of the calculation (no casting)
volatile uint8_t delay_time;
volatile uint8_t channel = 0;									// Channel number to be read successive one every interrupt
const uint16_t attack_target = 4095;					// Maximum output voltage
const uint8_t decay_target = 0;								// Minimum output voltage

																							// Adapt the three constants below to your needs or preferences
const uint16_t envelope_high_target = 4080;		// Upper limit for envelope. Affects the timing of attack high tail
																							// To shorten the high tail increase envelope_high_target
const uint8_t envelope_low_target = 20;				// Lower limit for envelope. Affects the timing of decay low tail
																							// To shorten the tail increase envelope_low_target
const uint8_t delay_loop_time = 4;						// Time in ms within the delay loop. 
																							// To increase delay time increase delay_loop_time


void init_timer0(void)						// Timer interrupt used for reading CV 
{
//	TCCR0B = (1<<CS01)|(1<<CS00);	// Timer clock prescaler /64
	TCCR0B = (1<<CS02)|(1<<CS00);		// Timer clock prescaler /1024 ; 20MHz:1024 = 19531,25Hz; fast enough!
	TIFR0 = 1<<TOV0;								// Set overflow flag
	TIMSK0 = 1<<TOIE0;							// Enable overflow interrupt
}

void init_ports(void)
{
	DDRD |= (1<<PD0);			// Output for delay signal and delay signal LED. High during delay.
	DDRD |= (1<<PD1);			// Output for attack signal and attack signal LED. High during attack.
	DDRD &=~ (1 << PD2);	// set PD2 as input; INT1  
  DDRD &=~ (1 << PD3);	// set PD3 as input; INT0  
	DDRD |= (1<<PD5);			// Output for decay signal and decay LED. High during decay.
	DDRD |= (1<<PD6);			// Output for end signal and EOC signal LED. High when the envelope is completed.
	DDRD &= ~(1 << PD7);	// PD7 Gate input
//	EIMSK |= (1<<INT0);			        // enable INT0 ; Uncomment those two lines if you want to use INT0
//  EICRA |= (1<<ISC00)|(1<<ISC01); // INT0 - Rising edge on INT0 generates an interrupt request	
//  EIMSK |= (1<<INT1);			        // enable INT1 ; Uncomment those two lines if you want to use INT1
//  EICRA |= (1<<ISC11);            // INT1 - Falling edge of INT1 generates interrupt request
}


void init_ADC() 
{
	
  uint16_t result;
 
  ADMUX |= (1 << REFS0);					// reference voltage on AVCC    
	ADMUX |= (0 << REFS1);					// reference voltage on AVCC
	ADMUX |= (1 << ADLAR);					// left adjust result, return only 8bits
  // Bit ADFR ("free running") in ADCSRA is already set to zero on startup; single conversion
  ADCSRA |= (1<<ADPS2);    				// Prescaler 20MHz / 16 = 1250kHz recommended: 50 - 200 kHz
																	// Faster conversion usable here because we use only 8bits
  ADCSRA |= (1<<ADEN);            // ADC activate
  ADCSRA |= (1<<ADSC);						// eine ADC-Wandlung; dummy conversion 
  while (ADCSRA & (1<<ADSC))			// Wait for end of conversion
	{         	
  }
  result = ADCW;									//ADCW must be read once (ignore compiler warning)
}

/* ADC single conversion */
uint8_t ADC_Read(uint8_t channel)
{
  ADMUX = (ADMUX & ~(0x1F)) | (channel & 0x1F);	// Select channel without interfering other bits
  ADCSRA |= (1<<ADSC);            							// "single conversion"
  while (ADCSRA & (1<<ADSC))										// Wait for completion
	{  
  }
  return (ADCH);																// Only 8bits are used        
}


void init_SPI(void)
	{
		DDRB |= (1<<PB5)|(1<<PB3)|(1<<PB2);			// SCK, MOSI and SS as outputs
    DDRB &= ~(1<<PB4);											// MISO as input
    SPCR |= (1<<MSTR);											// Set AVR m328p as Master
    SPCR |= (0<<SPR0)|(0<<SPR1);						// divide clock by 4
//		SPCR |= (1<<SPR0)|(0<<SPR1);						// divide clock by 16
//		SPCR |= (0<<SPR0)|(1<<SPR1);						// divide clock by 64
//		SPCR |= (1<<SPR0)|(1<<SPR1);						// divide clock by 128
		SPSR |= (1<<SPI2X);											// double SPI speed (SCK)
    SPCR |= (1<<SPE);												// Enable SPI
	}


void Transmit_SPI_Master(uint16_t Data)					// Send two bytes to MCP4921
{	
	PORTB &=~ (1<<PB2);    												// assert the slave select
	SPDR = ((Data>>8) & 0x0F)|(config_bits<<4);		// Start transmission, send high byte first
	while (!(SPSR & (1<<SPIF)));									// Wait for transmission complete
	SPDR = Data & 0xFF;														// send low byte next
	while (!(SPSR & (1<<SPIF)));									// Wait for transmission complete
	PORTB |= (1<<PB2);														// de-assert slave select
}


void update_inputs(uint8_t scan)								// read the CV input parameters. One every loop.
	{        
		uint8_t attack_CV;
		uint8_t decay_CV;
		
		switch (scan)
		{
			case 0:
				delay_time = ADC_Read(0);								// get the delay time CV
				if (delay_time < 3)											// Minimum delay time effective above 2ms
					delay_time = 0;
			break;
			case 1:
				attack_CV = ADC_Read(1);								// get the attack time CV
				attack_time = pgm_read_float_near(&(lookup_attack[attack_CV]));	// get the attack time increment
			break; 
			case 2:
				decay_CV = ADC_Read(2);									// get the decay time CV
				decay_time = pgm_read_float_near(&(lookup_decay[decay_CV]));		// get the decay time increment
			break; 
		}
  }

/*	Uncomment if you want to use INT0. Don't forget init_port()s.
ISR(INT0_vect)              // Gate in goes high
{
	Put your code here
}  
*/

/* Uncomment if you want to use INT1. Don't forget init_ports().
ISR(INT1_vect)              // Gate in goes low
{
//	Put your code here
	retrigger = true;
}  
*/

ISR (TIMER0_OVF_vect)									// (Did not work when variable channel is declared inside ISR)
	{
		update_inputs(channel);						// Update control voltage on timer interrupt
		channel = channel + 1;						// Select next channel
		if (channel > 2)									// All channels read?
			channel = 0;										// Start over
	}
	
	
int main(void)
	{
		float envelope = 0.0;
		bool gate = false;								// Holds gate state
		bool busy = false;								// True while envelope is active 
		uint8_t i;												// loop counter

		
		cli();
		init_ports();
		init_SPI();
		init_ADC();
		init_timer0();
		sei();
		Transmit_SPI_Master(0);						// Set output to zero on start
		PORTD |= (1 << PD6);							// PD6 EOC state high on power on
		
		while(1)
		{        		
			gate = (PIND & (1 << PD7));									// Read gate state every loop == no retrigger during loop
			if ((gate == true) && (busy == false))			// Initial condition for envelope start: Gate == high and ENV not in progress
			{
				busy = true;															// No retrigger before EOC
				PORTD &= ~(1 << PD6);											// PD6 EOC state off
				if (delay_time>2)													// Delay loop starts here
				{
					PORTD |= (1 << PD5);										// PD5 LED and delay state on
					for (i = 0; i<delay_time;i++)						// Delay loop for delay time
					 _delay_ms(delay_loop_time);						// Delay time = 255 x delay_loop_time in msec
					PORTD &= ~(1 << PD5);										// PD5 LED and delay state off
				}																					// Delay loop ends here
				PORTD |= (1 << PD1);											// PD1 LED and attack state on
				while (envelope < envelope_high_target)		// Termination condition for while:  Max ENV voltage
				{																					// Setting to 4000 cuts off the "long tail" 4080 gives more of a plateau in the top. 
					envelope=((1.0-attack_time)*attack_target+attack_time*envelope);	// implement difference equation: y(k) = (1 - alpha) * x(k) + alpha * y(k-1)
					Transmit_SPI_Master(round(envelope));		// Output current envelope voltage
				}																					// Attack reached maximum 
				PORTD &= ~(1 << PD1);											// PD1 LED and attack state off
																									// Release starts here
				PORTD |= (1 << PD0);											// PD0 LED and decay state on
				while (envelope > envelope_low_target)		// Termination condition for while: Minimum ENV voltage
																									// Setting to 80 cuts off the "long tail" 4 gives more of a tail in the bottom.
				{																					// But this gives a bottom voltage value of about 70mV.
					envelope=((1.0-decay_time)*decay_target+decay_time*envelope);	// implement difference equation: y(k) = (1 - alpha) * x(k) + alpha * y(k-1)
					Transmit_SPI_Master(round(envelope));		// Output current envelope voltage
				}																					// Decay reached minimum
				Transmit_SPI_Master(0);										// Forces the output to zero. Might be useful if you choose to cut the long tail on decay.
				PORTD &= ~(1 << PD0);											// PD0 LED and decay state off
				PORTD |= (1 << PD6);											// PD6 LED and EOC state on
			}																						// End of if gate = high
			
			if (gate == false)													// This point is reached after the ENV has reached the end
			{			
				busy = false;															// Ready for new ENV
			}								
		}  
		return(0);																		// You'll never reach this line
	}
